/**
 @name Create3DBox
 @version 2.5 (05 June 2009)
 @fileoverview
 <h4>Description</h4>
 <p>This script creates a 3D box with user-defined layers.
 Assign a layer to each side of the box, and the script will position and scale them to build a box of given size.
 Note that the same layer can be used for multiple sides (the script will create copies of the layer).
 This allows to create a box with only one layer in the composition.
 For easy control, the "Add Controller" option parents each side layer to a 3D Null positioned at the center of the box.</p>
 <h4>Usage</h4>
 <ol>
    <li> In After Effects CS3 or later, run the script 
    <li> Open a comp containing at least one layer
    <li> Click on "Refresh" to update available layers list
    <li> Set the dimension of the box, and assign a layer to each side of the box
    <li> Click on "Create" to build the box
 </ol>
 @author Charles Bordenave
 @location http://www.nabscripts.com/downloads/scripts/Create3DBox.zip
*/


/**
 This class represents the main class of the script;
 it is used to create the user interface which allows to create a 3D box
 @class Main class of the script
*/
function Create3DBox()
{
    // Variable used to keep track of 'this' reference
    var create3DBox = this;

    // Create an instance of the utils class to use its functions
    var utils = new Create3DBoxUtils();

    // Script infos
    this.scriptMinSupportVersion = "8.0";
    this.scriptName = "Create3DBox.jsx";
    this.scriptVersion = "2.5";
    this.scriptTitle = "Create 3D Box";
    this.scriptCopyright = "Copyright (c) 2009 Charles Bordenave";
    this.scriptHomepage = "http://www.nabscripts.com";
    this.scriptDescription = {en: "This script creates a 3D box with user-defined layers.\\r\\rAssign a layer to each side of the box, and the script will position and scale them to build a box of given size.\\r\\rNote that the same layer can be used for multiple sides (the script will create copies of the layer). This allows to create a box with only one layer in the composition.\\r\\rFor easy control, the \"Add Controller\" option parents each side layer to a 3D Null positioned at the center of the box.", fr:"Ce script crée une boîte 3D avec les calques choisis par l\\'utilisateur.\\r\\rAssignez un calque à chacune des faces de la boîte et le script va les positionner et les redimensionner pour construire une boîte d\\'une taille donnée.\\r\\rNotez que le même calque peut servir pour plusieurs faces (le script fera des copies du calque). Ceci permet par exemple de créer une boîte avec un seul calque dans la composition.\\r\\rPour un meilleur contrôle, l\\'option \"Ajouter Contrôleur\" parente chaque face à un Nul 3D positionné au centre de la boîte."};
    this.scriptAbout = {en:this.scriptName + ", v" + this.scriptVersion + "\\r\\r" + this.scriptCopyright + "\\r" + this.scriptHomepage + "\\r\\r" + utils.loc(this.scriptDescription), fr:this.scriptName + ", v" + this.scriptVersion + "\\r\\r" + this.scriptCopyright + "\\r" + this.scriptHomepage + "\\r\\r" + utils.loc(this.scriptDescription)};

    // Errors
    this.requirementErr = {en:"This script requires After Effects CS3 or later.", fr:"Ce script nécessite After Effects CS3 ou supérieur."};
    this.noCompErr = {en:"A comp must be active.", fr:"Une composition doit être active."};
    this.noLayersErr = {en:"The comp must contain at least one layer.", fr:"La composition doit contenir au moins un calque."};
    this.noSelErr = {en:"You must assign a layer to each side of the box. Do not forget to Refresh layers list if you have opened a new comp or created new layers.", fr:"Vous devez assigner un calque à chacune des faces de la boîte. N'oubliez pas d'Actualiser la liste des calques si vous avez ouvert une nouvelle comp ou créé de nouveaux calques."};
    this.badLayersErr = {en:"Layer \"%s\" cannot be used as a side of the box.", fr:"Le calque \"%s\" ne peut pas être utilisé pour créer la boîte."};

    // UI strings & default settings
    this.aboutBtnName = "?";
    this.dimensionsPnlName = {en:"Dimensions", fr:"Dimensions"};
    this.widthStName = {en:"Width:", fr:"Largeur:"};
    this.heightStName = {en:"Height:", fr:"Hauteur:"};
    this.depthStName = {en:"Depth:", fr:"Profondeur:"};
    this.uniformCbName = {en:"Uniform Size", fr:"Taille uniforme"};
    this.widthEtDflt = 100;
    this.heightEtDflt = 100;
    this.depthEtDflt = 100;
    this.texturesPnlName = {en:"Textures", fr:"Textures"};
    this.refreshBtnName = {en:"Refresh", fr:"Actualiser"};
    this.listWidth = 100; // preferredSize.width of dropdownlists
    this.frontStName = {en:"Front:", fr:"Avant:"};
    this.backStName = {en:"Back:", fr:"Arrière:"};
    this.leftStName = {en:"Left:", fr:"Gauche:"};
    this.rightStName = {en:"Right:", fr:"Droite:"};
    this.bottomStName = {en:"Bottom:", fr:"Bas:"};
    this.topStName = {en:"Top:", fr:"Haut:"};
    this.addControllerCbName = {en:"Add Controller", fr:"Ajouter Contrôleur"};
    this.controllerLayerName = {en:"Controller", fr:"Contrôleur"};
    this.runBtnName = {en:"Create", fr:"Créer"};

    // Miscellaneous 
    this.nameSep = " * "; 
    this.layerNames = {en:["Front","Back","Left","Right","Bottom","Top"], fr:["Avant","Arrière","Gauche","Droite","Bas","Haut"]}; // "LayerName" ==> "Back * LayerName"

    /**
     Creates and displays the script interface
     @param {Object} thisObj A Panel object if the script is launched from the Window menu, null otherwise
    */
    this.buildUI = function (thisObj)
    {
        // dockable panel or palette
        var pal = (thisObj instanceof Panel) ? thisObj : new Window("palette", this.scriptTitle, undefined, {resizeable:false});

        // resource specifications
        var res =
        "group { orientation:'column', alignment:['left','top'], alignChildren:'fill', \
            gr1: Group { \
                aboutBtn: Button { text:'" + this.aboutBtnName + "', preferredSize:[25,20], alignment:['right','center'] } \
            }, \
            gr2: Panel { orientation:'row', alignment:['fill','fill'], alignChildren:['center','center'], text:'" + utils.loc(this.dimensionsPnlName) + "', \
                widthSt: StaticText { text:'" + utils.loc(this.widthStName) + "' }, \
                widthEt: EditText { text:'" + this.widthEtDflt + "', characters:5 }, \
                heightSt: StaticText { text:'" + utils.loc(this.heightStName) + "', enabled:false }, \
                heightEt: EditText { text:'" + this.heightEtDflt + "', characters:5, enabled:false }, \
                depthSt: StaticText { text:'" + utils.loc(this.depthStName) + "', enabled:false }, \
                depthEt: EditText { text:'" + this.depthEtDflt + "', characters:5, enabled:false }, \
                uniformCb: Checkbox { text:'" + utils.loc(this.uniformCbName) + "', value:true } \
            }, \
            gr3: Panel { orientation:'row', alignment:['fill','center'], text:'" + utils.loc(this.texturesPnlName) + "', \
                gr31: Group { alignment:['center','top'], \
                    refreshBtn: Button { text:'" + utils.loc(this.refreshBtnName) + "' } \
                }, \
                gr32: Group { orientation:'row', \
                    gr321: Group { orientation:'column', alignChildren:['right','top'], \
                        gr3211: Group { orientation:'row', \
                            frontSt: StaticText { text:'" + utils.loc(this.frontStName) + "' }, \
                            frontLst: DropDownList { preferredSize:['" + this.listWidth + "',20] } \
                        }, \
                        gr3212: Group { orientation:'row', \
                            leftSt: StaticText { text:'" + utils.loc(this.leftStName) + "' }, \
                            leftLst: DropDownList { preferredSize:['" + this.listWidth + "',20] } \
                        }, \
                        gr3213: Group { orientation:'row', \
                            bottomSt: StaticText { text:'" + utils.loc(this.bottomStName) + "' }, \
                            bottomLst: DropDownList { preferredSize:['" + this.listWidth + "',20] } \
                        } \
                    }, \
                    gr322: Group { orientation:'column', alignChildren:['right','top'], \
                        gr3221: Group { orientation:'row', \
                            backSt: StaticText { text:'" + utils.loc(this.backStName) + "' }, \
                              backLst: DropDownList { preferredSize:['" + this.listWidth + "',20] } \
                        }, \
                        gr3222: Group { orientation:'row', \
                            rightSt: StaticText { text:'" + utils.loc(this.rightStName) + "' }, \
                            rightLst: DropDownList { preferredSize:['" + this.listWidth + "',20] } \
                        }, \
                        gr3223: Group { orientation:'row', \
                            topSt: StaticText { text:'" + utils.loc(this.topStName) + "' }, \
                            topLst: DropDownList { preferredSize:['" + this.listWidth + "',20] } \
                        } \
                    } \
                }, \
            }, \
            gr4: Group { orientation:'row', alignment:['fill','top'], \
                addControllerCb: Checkbox { text:'" + utils.loc(this.addControllerCbName) + "', alignment:['left','center'], value:true }, \
                runBtn: Button { text:'" + utils.loc(this.runBtnName) + "', alignment:['right','center'] } \
            } \
        }";
        pal.gr = pal.add(res);

        pal.gr.gr2.widthEt.graphics.foregroundColor = 
        pal.gr.gr2.heightEt.graphics.foregroundColor = 
        pal.gr.gr2.depthEt.graphics.foregroundColor =         
        pal.gr.gr3.gr32.gr321.gr3211.frontLst.graphics.foregroundColor = 
        pal.gr.gr3.gr32.gr321.gr3212.leftLst.graphics.foregroundColor = 
        pal.gr.gr3.gr32.gr321.gr3213.bottomLst.graphics.foregroundColor = 
        pal.gr.gr3.gr32.gr322.gr3221.backLst.graphics.foregroundColor = 
        pal.gr.gr3.gr32.gr322.gr3222.rightLst.graphics.foregroundColor = 
        pal.gr.gr3.gr32.gr322.gr3223.topLst.graphics.foregroundColor = pal.graphics.newPen(pal.graphics.BrushType.SOLID_COLOR, [0,0,0], 1);
       
        // event callbacks
        pal.gr.gr1.aboutBtn.onClick = function ()
        {
            utils.createAboutDlg(create3DBox.scriptAbout);
        };

        pal.gr.gr2.widthEt.onChange = pal.gr.gr2.widthEt.onChanging = function ()
        {
            if (isNaN(this.text) || parseFloat(this.text) < 1) this.text = create3DBox.widthEtDflt;
            this.text = parseFloat(this.text);

            if (pal.gr.gr2.uniformCb.value)
            {
                pal.gr.gr2.heightEt.text = pal.gr.gr2.depthEt.text = this.text;
            }
        };

        pal.gr.gr2.uniformCb.onClick = function ()
        {
            if (this.value)
            {
                pal.gr.gr2.heightEt.text = pal.gr.gr2.depthEt.text = pal.gr.gr2.widthEt.text;
            }
            pal.gr.gr2.heightSt.enabled = pal.gr.gr2.depthSt.enabled =
            pal.gr.gr2.heightEt.enabled = pal.gr.gr2.depthEt.enabled = !this.value;
        };

        pal.gr.gr3.gr31.refreshBtn.onClick = function ()
        {
            try
            {
                var comp = app.project.activeItem;
                var err = create3DBox.noCompErr;
                if (create3DBox.checkActiveItem(comp)) throw(err);

                var lists = [
                pal.gr.gr3.gr32.gr321.gr3211.frontLst,
                pal.gr.gr3.gr32.gr321.gr3212.leftLst,
                pal.gr.gr3.gr32.gr321.gr3213.bottomLst,
                pal.gr.gr3.gr32.gr322.gr3221.backLst,
                pal.gr.gr3.gr32.gr322.gr3222.rightLst,
                pal.gr.gr3.gr32.gr322.gr3223.topLst
                ];

                for (var i = 1; i <= comp.numLayers; i++)
                {
                    for (var j = 0; j < lists.length; j++)
                    {
                        if (i == 1) lists[j].removeAll();

                        lists[j].add("item", comp.layer(i).name);

                        if (i == 1) lists[j].selection = 0;
                    }
                }
            }
            catch(err)
            {
                utils.throwErr(err);
            }
        };

        pal.gr.gr4.runBtn.onClick = function ()
        {
            create3DBox.createBox(pal);
        };

        // show user interface
        if (pal instanceof Window)
        {
            pal.center();
            pal.show();
        }
        else
        {
            pal.layout.layout(true);
        }
    };

    /**
     Determines whether the active item is a composition
     @return True if the active item is not a composition, False otherwise
    */
    this.checkActiveItem = function ()
    {
        var err = false;
        var comp = app.project.activeItem;
        if (!comp || !(comp instanceof CompItem))
        {
            err = true;
        }
        return err;
    };

    /**
     Creates a 3D Null and make it parent of the layers passed as arguments
     @param {Array} layers An array of layers to be parented to the controller
    */
    this.addController = function (layers, p)//os)
    {
        var comp = layers[0].containingComp;
        var controllerLayer = comp.layers.addNull();
        controllerLayer.name = utils.loc(this.controllerLayerName);
        controllerLayer.threeDLayer = true;
        controllerLayer.enabled = false;

        var pos = [0,0,0];
        for (var i = 0; i < layers.length; i++) pos += layers[i].position.valueAtTime(comp.time, false);
        pos /= layers.length;

        var p = controllerLayer.position;
        p.numKeys ? p.setValueAtTime(comp.time, pos) : p.setValue(pos);

        for (var i = 0; i < layers.length; i++) layers[i].parent = controllerLayer;

        controllerLayer.position.setValue([comp.width/2,comp.height/2,0]);
    };

    /**
     Scales a layer so that its size (in pixels) matches the desired size
     @param {Object} layer A layer object to resize
     @param {Array} desiredSize A two-dimensional array representing the width and height of the desired size
     @param {Boolean} parFlag A flag that indicates whether the layer requires special care about comp pixel aspect ratio
    */
    this.resizeToFit = function (layer, desiredSize, parFlag)
    {
        var compPar = layer.containingComp.pixelAspect;
        var layerPar = layer.source ? layer.source.pixelAspect : 1.0;
        
        var layerSize = utils.getLayerSize(layer, layer.containingComp.time);
        
        var layerW = layerSize[0] * layerPar; // real width
        var layerH = layerSize[1];        

        var targetW = desiredSize[0] * Math.pow(layerPar, Number(parFlag)+1) / layerPar; // same as   parFlag ? desiredSize[0] * layerPar : desiredSize[0];
        if (layerPar != 1.0 && layerPar != compPar && parFlag) targetW *= compPar/layerPar;
        
        var targetH = desiredSize[1];

        var sx = 100 * targetW / layerW;
        var sy = 100 * targetH / layerH;

        var s = layer.scale;
        s.numKeys ? s.setValueAtTime(layer.containingComp.time, [sx,sy]) : s.setValue([sx,sy]);
    };

    /**
     Functional part of the script: creates a 3D box according to user settings
     @param {Object} pal A palette or a dockable panel containing all user parameters
    */
    this.createBox = function (pal)
    {
        try
        {
            var comp = app.project.activeItem;
            var err = this.noCompErr;
            if (this.checkActiveItem(comp)) throw(err);
            
            var err = this.noLayersErr;
            if (comp.layers.length < 1) throw(err);
            
            var err = this.noSelErr;
            try
            {
                var selfront = comp.layer(pal.gr.gr3.gr32.gr321.gr3211.frontLst.selection);//.index + 1);
                var selback = comp.layer(pal.gr.gr3.gr32.gr322.gr3221.backLst.selection);//.index + 1);
                var selleft = comp.layer(pal.gr.gr3.gr32.gr321.gr3212.leftLst.selection);//.index + 1);
                var selright = comp.layer(pal.gr.gr3.gr32.gr322.gr3222.rightLst.selection);//.index + 1);
                var selbottom = comp.layer(pal.gr.gr3.gr32.gr321.gr3213.bottomLst.selection);//.index + 1);
                var seltop = comp.layer(pal.gr.gr3.gr32.gr322.gr3223.topLst.selection);//.index + 1);
                if (!selfront || !selback || !selleft || !selright || !selbottom || !seltop) throw(err);
            }
            catch(e)
            {
                throw(err);
            }

            // initialization
            var front = selfront;
            var back = selback;
            var left = selleft;
            var right = selright;
            var bottom = selbottom;
            var top = seltop;

            // box settings
            var boxW = parseFloat(pal.gr.gr2.widthEt.text) / comp.pixelAspect;
            var boxH = parseFloat(pal.gr.gr2.heightEt.text);
            var boxD = parseFloat(pal.gr.gr2.depthEt.text);
            var halfW = 0.5 * boxW;
            var halfH = 0.5 * boxH;
            var halfD = 0.5 * boxD;
            var pos = [[halfW, halfH, 0], [halfW, halfH, boxD], [0, halfH, halfD], [boxW, halfH, halfD], [halfW, boxH, halfD], [halfW, 0, halfD]];
            var ori = [[0,0,0], [0,180,0], [0,90,0], [0,270,0], [90,0,0], [270,0,0]];
        
            app.beginUndoGroup(this.scriptTitle);

            // create duplicates if necessary
            if (selback == selfront) back = selback.duplicate();
            if (selleft == selfront || selleft == selback) left = selleft.duplicate();
            if (selright == selfront || selright == selback || selright == selleft) right = selright.duplicate();
            if (selbottom == selfront || selbottom == selback || selbottom == selleft || selbottom == selright) bottom = selbottom.duplicate();
            if (seltop == selfront || seltop == selback || seltop == selleft || seltop == selright || seltop == selbottom) top = seltop.duplicate();

            var layers = [front,back,left,right,bottom,top];

            // create the box
            var err = this.badLayersErr;
            try
            {
                for (var i = 0; i < layers.length; i++)
                {
                    layers[i].threeDLayer = true;
                    var layerPar = layers[i].source ? layers[i].source.pixelAspect : 1.0;
                    var layerSize = utils.getLayerSize(layers[i], comp.time);
                    var layerW = layerSize[0]; 
                    var layerH = layerSize[1];
                    var a = layers[i].anchorPoint;
                    var p = layers[i].position;
                    var r = layers[i].orientation;
                    
                    if (layers[i] instanceof TextLayer) // assume left justify, since cs3 does not support font stuff
                        a.numKeys ? a.setValueAtTime(comp.time, 0.5 * [layerW, -layerH, 0]) : a.setValue(0.5 * [layerW, -layerH, 0]);
                    else
                        a.numKeys ? a.setValueAtTime(comp.time, 0.5 * [layerW, layerH, 0]) : a.setValue(0.5 * [layerW, layerH, 0]);
                    
                    if (layerPar == 1.0)
                    {
                        var pp = [pos[i][0]/comp.pixelAspect, pos[i][1], pos[i][2]];
                        p.numKeys ? p.setValueAtTime(comp.time, pp) : p.setValue(pp);    
                    }
                    else
                        p.numKeys ? p.setValueAtTime(comp.time, pos[i]) : p.setValue(pos[i]);
                    r.numKeys ? r.setValueAtTime(comp.time, ori[i]) : r.setValue(ori[i]); 
                    
                    if (i == 0 || i == 1) this.resizeToFit(layers[i],[boxW,boxH], true); // front & back
                    if (i == 2 || i == 3) this.resizeToFit(layers[i],[boxD,boxH], false); // left & right
                    if (i == 4 || i == 5) this.resizeToFit(layers[i],[boxW,boxD], true); // bottom & top
                }
            }
            catch(e)
            {
                err.en = err.en.replace('%s',layers[i].name);
                err.fr = err.fr.replace('%s',layers[i].name);
                throw(err);
            }

            // add controller if requested
            if (pal.gr.gr4.addControllerCb.value)
            {
                this.addController(layers, [halfW, -halfH, halfD]);
            }

            // rename layers (maybe optional...)                    
            for (var i = 0; i < layers.length; i++)
            {
                layers[i].name = utils.loc(this.layerNames)[i] + this.nameSep + layers[i].name.substring(0,31 - this.nameSep.length - utils.loc(this.layerNames)[i].length);
            }

            app.endUndoGroup();
        }
        catch(err)
        {
            utils.throwErr(err);
        }
    };

    /**
     Runs the script
     @param {Object} thisObj A Panel object if the script is launched from the Window menu, null otherwise
    */
    this.run = function (thisObj)
    {
        if (parseFloat(app.version) < parseFloat(this.scriptMinSupportVersion))
        {
            this.throwErr(this.requirementErr);
        }
        else
        {
            this.buildUI(thisObj);
        }
    };
}


/**
 This class provides some utility functions used by Create3DBox
 @class Some utility functions grouped in a class
*/
function Create3DBoxUtils()
{
    /**
     String localization function: english and french languages are supported
     @param {Object} str A localization object containing the localized versions of a string
     @return Appropriate localized version of str
    */
    this.loc = function (str)
    {
        return app.language == Language.FRENCH ? str.fr : str.en;
    };

    /**
     Displays a window containg a localized error message
     @param {Object} err A localization object containing the localized versions of an error message
    */
    this.throwErr = function (err)
    {
        var wndTitle = $.fileName.substring($.fileName.lastIndexOf("/")+1, $.fileName.lastIndexOf("."));
        Window.alert("Script error:\r" + this.loc(err), wndTitle, true);
    };

    /**
     Get the size in pixels of an AV layer (comp layer, footage layer, solid layer, text layer) at specific time
     @param {Object} avLayer An AV layer object    
     @param {Number} time A floating-point value representing the time in seconds at which the layer size must be evaluated    
     @return A two-dimensional array containing the width and height of the layer
    */
    this.getLayerSize = function (avLayer, time)
    {
        var w, h;
        if (!(avLayer instanceof TextLayer))
        {
            w = avLayer.width;
            h = avLayer.height;
        }
        else
        {
            var bb = avLayer.sourceRectAtTime(time, true);
            w = bb.width;
            h = bb.height;
        }
        return [w,h];
    };
    
    /**
     Displays a customized window containg the About text
     @param {String} aboutStr The text to display
    */
    this.createAboutDlg = function (aboutStr)
    {
        eval(unescape('%20%20%20%20%20%20%20%20%2F%2A%2A%20%0A%20%20%20%20%20%20%20%20%20%44%72%61%77%20%73%6F%6D%65%20%72%61%6E%64%6F%6D%20%72%65%63%74%61%6E%67%6C%65%73%20%28%70%6F%73%69%74%69%6F%6E%2C%20%73%69%7A%65%2C%20%63%6F%6C%6F%72%2C%20%61%6C%70%68%61%29%20%6F%6E%20%74%68%65%20%77%69%6E%64%6F%77%20%62%61%63%6B%67%72%6F%75%6E%64%0A%20%20%20%20%20%20%20%20%20%40%69%67%6E%6F%72%65%20%0A%20%20%20%20%20%20%20%20%2A%2F%0A%20%20%20%20%20%20%20%20%66%75%6E%63%74%69%6F%6E%20%61%64%64%4E%61%62%73%63%72%69%70%74%73%42%61%63%6B%67%72%6F%75%6E%64%53%69%67%6E%61%74%75%72%65%28%77%6E%64%29%0A%20%20%20%20%20%20%20%20%7B%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%6E%75%6D%52%65%63%74%20%3D%20%32%34%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%6D%69%6E%4F%70%61%63%69%74%79%20%3D%20%30%2E%30%35%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%6D%61%78%4F%70%61%63%69%74%79%20%3D%20%30%2E%31%35%3B%0A%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%6C%65%66%74%45%64%67%65%20%3D%20%30%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%74%6F%70%45%64%67%65%20%3D%20%30%3B%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%72%69%67%68%74%45%64%67%65%20%3D%20%77%6E%64%2E%77%69%6E%64%6F%77%42%6F%75%6E%64%73%2E%77%69%64%74%68%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%62%6F%74%74%6F%6D%45%64%67%65%20%3D%20%77%6E%64%2E%77%69%6E%64%6F%77%42%6F%75%6E%64%73%2E%68%65%69%67%68%74%3B%0A%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%66%6F%72%20%28%76%61%72%20%69%20%3D%20%30%20%3B%20%69%20%3C%20%6E%75%6D%52%65%63%74%3B%20%69%2B%2B%29%0A%20%20%20%20%20%20%20%20%20%20%20%20%7B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%78%4C%6F%63%20%3D%20%31%30%20%2B%20%28%72%69%67%68%74%45%64%67%65%20%2D%20%32%30%29%20%2A%20%4D%61%74%68%2E%72%61%6E%64%6F%6D%28%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%79%4C%6F%63%20%3D%20%31%30%20%2B%20%28%62%6F%74%74%6F%6D%45%64%67%65%20%2D%20%32%30%29%20%2A%20%4D%61%74%68%2E%72%61%6E%64%6F%6D%28%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%77%69%64%74%68%20%3D%20%35%20%2B%20%31%35%20%2A%20%4D%61%74%68%2E%72%61%6E%64%6F%6D%28%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%68%65%69%67%68%74%20%3D%20%35%20%2B%20%31%35%20%2A%20%4D%61%74%68%2E%72%61%6E%64%6F%6D%28%29%3B%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%62%6F%72%64%65%72%57%69%64%74%68%20%3D%20%31%20%2B%20%34%20%2A%20%4D%61%74%68%2E%72%61%6E%64%6F%6D%28%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%62%6F%72%64%65%72%43%6F%6C%6F%72%20%3D%20%5B%4D%61%74%68%2E%72%61%6E%64%6F%6D%28%29%2C%20%4D%61%74%68%2E%72%61%6E%64%6F%6D%28%29%2C%20%4D%61%74%68%2E%72%61%6E%64%6F%6D%28%29%2C%20%6D%69%6E%4F%70%61%63%69%74%79%20%2B%20%28%6D%61%78%4F%70%61%63%69%74%79%20%2D%20%6D%69%6E%4F%70%61%63%69%74%79%29%20%2A%20%4D%61%74%68%2E%72%61%6E%64%6F%6D%28%29%5D%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%63%6F%6C%6F%72%42%72%75%73%68%20%3D%20%77%6E%64%2E%67%72%61%70%68%69%63%73%2E%6E%65%77%42%72%75%73%68%28%77%6E%64%2E%67%72%61%70%68%69%63%73%2E%42%72%75%73%68%54%79%70%65%2E%53%4F%4C%49%44%5F%43%4F%4C%4F%52%2C%20%62%6F%72%64%65%72%43%6F%6C%6F%72%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%77%6E%64%2E%67%31%20%3D%20%77%6E%64%2E%61%64%64%28%22%67%72%6F%75%70%22%2C%20%5B%78%4C%6F%63%2C%20%79%4C%6F%63%2C%20%78%4C%6F%63%20%2B%20%77%69%64%74%68%2C%20%79%4C%6F%63%20%2B%20%62%6F%72%64%65%72%57%69%64%74%68%5D%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%77%6E%64%2E%67%32%20%3D%20%77%6E%64%2E%61%64%64%28%22%67%72%6F%75%70%22%2C%20%5B%78%4C%6F%63%2C%20%79%4C%6F%63%20%2B%20%68%65%69%67%68%74%20%2D%20%62%6F%72%64%65%72%57%69%64%74%68%2C%20%78%4C%6F%63%20%2B%20%77%69%64%74%68%2C%20%79%4C%6F%63%20%2B%20%68%65%69%67%68%74%5D%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%77%6E%64%2E%67%33%20%3D%20%77%6E%64%2E%61%64%64%28%22%67%72%6F%75%70%22%2C%20%5B%78%4C%6F%63%2C%20%79%4C%6F%63%20%2B%20%62%6F%72%64%65%72%57%69%64%74%68%2C%20%78%4C%6F%63%20%2B%20%62%6F%72%64%65%72%57%69%64%74%68%2C%20%79%4C%6F%63%20%2B%20%68%65%69%67%68%74%20%2D%20%62%6F%72%64%65%72%57%69%64%74%68%5D%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%77%6E%64%2E%67%34%20%3D%20%77%6E%64%2E%61%64%64%28%22%67%72%6F%75%70%22%2C%20%5B%78%4C%6F%63%20%2B%20%77%69%64%74%68%20%2D%20%62%6F%72%64%65%72%57%69%64%74%68%2C%20%79%4C%6F%63%20%2B%20%62%6F%72%64%65%72%57%69%64%74%68%2C%20%78%4C%6F%63%20%2B%20%77%69%64%74%68%2C%20%79%4C%6F%63%20%2B%20%68%65%69%67%68%74%20%2D%20%62%6F%72%64%65%72%57%69%64%74%68%5D%29%3B%0A%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%77%6E%64%2E%67%31%2E%67%72%61%70%68%69%63%73%2E%62%61%63%6B%67%72%6F%75%6E%64%43%6F%6C%6F%72%20%3D%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%77%6E%64%2E%67%32%2E%67%72%61%70%68%69%63%73%2E%62%61%63%6B%67%72%6F%75%6E%64%43%6F%6C%6F%72%20%3D%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%77%6E%64%2E%67%33%2E%67%72%61%70%68%69%63%73%2E%62%61%63%6B%67%72%6F%75%6E%64%43%6F%6C%6F%72%20%3D%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%77%6E%64%2E%67%34%2E%67%72%61%70%68%69%63%73%2E%62%61%63%6B%67%72%6F%75%6E%64%43%6F%6C%6F%72%20%3D%20%63%6F%6C%6F%72%42%72%75%73%68%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%7D%20%20%20%20%0A%20%20%20%20%20%20%20%20%7D%0A%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%76%61%72%20%64%6C%67%20%3D%20%6E%65%77%20%57%69%6E%64%6F%77%28%22%64%69%61%6C%6F%67%22%2C%20%22%41%62%6F%75%74%22%29%3B%0A%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%2F%2F%20%70%61%6E%65%6C%20%62%6F%72%64%65%72%53%74%79%6C%65%3A%20%6F%6E%65%20%6F%66%20%62%6C%61%63%6B%2C%20%65%74%63%68%65%64%2C%20%67%72%61%79%2C%20%72%61%69%73%65%64%2C%20%73%75%6E%6B%65%6E%2E%20%44%65%66%61%75%6C%74%20%69%73%20%65%74%63%68%65%64%2E%0A%20%20%20%20%20%20%20%20%2F%2F%20%72%65%73%6F%75%72%63%65%20%73%70%65%63%69%66%69%63%61%74%69%6F%6E%73%0A%20%20%20%20%20%20%20%20%76%61%72%20%72%65%73%20%3D%0A%20%20%20%20%20%20%20%20%22%67%72%6F%75%70%20%7B%20%6F%72%69%65%6E%74%61%74%69%6F%6E%3A%27%63%6F%6C%75%6D%6E%27%2C%20%61%6C%69%67%6E%6D%65%6E%74%3A%5B%27%66%69%6C%6C%27%2C%27%66%69%6C%6C%27%5D%2C%20%5C%0A%20%20%20%20%20%20%20%20%20%20%20%20%61%62%6F%75%74%50%6E%6C%3A%20%50%61%6E%65%6C%20%7B%20%70%72%6F%70%65%72%74%69%65%73%3A%7B%20%62%6F%72%64%65%72%53%74%79%6C%65%3A%27%73%75%6E%6B%65%6E%27%20%7D%2C%20%5C%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%61%62%6F%75%74%45%74%3A%20%45%64%69%74%54%65%78%74%20%7B%20%74%65%78%74%3A%27%22%20%2B%20%74%68%69%73%2E%6C%6F%63%28%61%62%6F%75%74%53%74%72%29%20%2B%20%22%27%2C%20%70%72%6F%70%65%72%74%69%65%73%3A%7B%6D%75%6C%74%69%6C%69%6E%65%3A%74%72%75%65%7D%2C%20%70%72%65%66%65%72%72%65%64%53%69%7A%65%3A%5B%32%38%30%2C%31%35%30%5D%2C%20%61%6C%69%67%6E%6D%65%6E%74%3A%5B%27%72%69%67%68%74%27%2C%27%63%65%6E%74%65%72%27%5D%20%7D%20%5C%0A%20%20%20%20%20%20%20%20%20%20%20%20%7D%2C%20%5C%0A%20%20%20%20%20%20%20%20%20%20%20%20%62%74%6E%73%47%72%3A%20%47%72%6F%75%70%20%7B%20%61%6C%69%67%6E%6D%65%6E%74%3A%5B%27%66%69%6C%6C%27%2C%27%66%69%6C%6C%27%5D%2C%20%5C%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%76%69%73%69%74%42%74%6E%3A%20%42%75%74%74%6F%6E%20%7B%20%74%65%78%74%3A%27%56%69%73%69%74%20%48%6F%6D%65%70%61%67%65%27%2C%20%61%6C%69%67%6E%6D%65%6E%74%3A%5B%27%6C%65%66%74%27%2C%27%63%65%6E%74%65%72%27%5D%20%7D%2C%20%5C%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%6F%6B%42%74%6E%3A%20%42%75%74%74%6F%6E%20%7B%20%74%65%78%74%3A%27%4F%6B%27%2C%20%61%6C%69%67%6E%6D%65%6E%74%3A%5B%27%72%69%67%68%74%27%2C%27%63%65%6E%74%65%72%27%5D%20%7D%20%5C%0A%20%20%20%20%20%20%20%20%20%20%20%20%7D%20%5C%0A%20%20%20%20%20%20%20%20%7D%22%3B%20%0A%20%20%20%20%20%20%20%20%64%6C%67%2E%67%72%20%3D%20%64%6C%67%2E%61%64%64%28%72%65%73%29%3B%0A%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%2F%2F%20%6F%6E%20%4D%61%63%20%77%65%20%63%61%6E%20%64%69%73%61%62%6C%65%20%65%64%69%74%20%74%65%78%74%20%77%68%69%6C%65%20%61%6C%6C%6F%77%69%6E%67%20%73%63%72%6F%6C%6C%69%6E%67%2C%20%6F%6E%20%57%69%6E%64%6F%77%73%20%77%65%20%63%61%6E%27%74%0A%20%20%20%20%20%20%20%20%64%6C%67%2E%67%72%2E%61%62%6F%75%74%50%6E%6C%2E%61%62%6F%75%74%45%74%2E%65%6E%61%62%6C%65%64%20%3D%20%28%24%2E%6F%73%2E%69%6E%64%65%78%4F%66%28%22%57%69%6E%22%29%20%21%3D%20%2D%31%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%2F%2F%20%64%72%61%77%20%72%61%6E%64%6F%6D%20%62%61%63%6B%67%72%6F%75%6E%64%20%63%6F%6C%6F%72%20%28%67%72%61%79%73%63%61%6C%65%29%0A%20%20%20%20%20%20%20%20%69%66%20%28%70%61%72%73%65%46%6C%6F%61%74%28%61%70%70%2E%76%65%72%73%69%6F%6E%29%20%3E%3D%20%39%2E%30%29%20%2F%2F%20%43%53%34%20%6F%72%20%6C%61%74%65%72%0A%20%20%20%20%20%20%20%20%7B%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%77%68%69%74%65%42%72%75%73%68%20%3D%20%64%6C%67%2E%67%72%61%70%68%69%63%73%2E%6E%65%77%42%72%75%73%68%28%64%6C%67%2E%67%72%61%70%68%69%63%73%2E%42%72%75%73%68%54%79%70%65%2E%53%4F%4C%49%44%5F%43%4F%4C%4F%52%2C%20%5B%31%2C%20%31%2C%20%31%2C%20%31%5D%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%72%61%6E%64%20%3D%20%4D%61%74%68%2E%72%61%6E%64%6F%6D%28%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%62%67%42%72%75%73%68%20%3D%20%64%6C%67%2E%67%72%61%70%68%69%63%73%2E%6E%65%77%42%72%75%73%68%28%64%6C%67%2E%67%72%61%70%68%69%63%73%2E%42%72%75%73%68%54%79%70%65%2E%53%4F%4C%49%44%5F%43%4F%4C%4F%52%2C%20%5B%72%61%6E%64%2C%20%72%61%6E%64%2C%20%72%61%6E%64%2C%20%31%5D%29%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%64%6C%67%2E%67%72%61%70%68%69%63%73%2E%62%61%63%6B%67%72%6F%75%6E%64%43%6F%6C%6F%72%20%3D%20%62%67%42%72%75%73%68%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%64%6C%67%2E%67%72%2E%61%62%6F%75%74%50%6E%6C%2E%67%72%61%70%68%69%63%73%2E%62%61%63%6B%67%72%6F%75%6E%64%43%6F%6C%6F%72%20%3D%20%77%68%69%74%65%42%72%75%73%68%3B%0A%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%64%6C%67%2E%6C%61%79%6F%75%74%2E%6C%61%79%6F%75%74%28%74%72%75%65%29%3B%20%2F%2F%20%74%6F%20%67%65%74%20%77%69%6E%64%6F%77%20%62%6F%75%6E%64%73%0A%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%61%64%64%4E%61%62%73%63%72%69%70%74%73%42%61%63%6B%67%72%6F%75%6E%64%53%69%67%6E%61%74%75%72%65%28%64%6C%67%29%3B%0A%20%20%20%20%20%20%20%20%7D%0A%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%64%6C%67%2E%67%72%2E%62%74%6E%73%47%72%2E%6F%6B%42%74%6E%2E%6F%6E%43%6C%69%63%6B%20%3D%20%66%75%6E%63%74%69%6F%6E%20%28%29%20%7B%20%64%6C%67%2E%63%6C%6F%73%65%28%29%3B%20%7D%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%2F%2F%20%6F%70%65%6E%20%68%6F%6D%65%70%61%67%65%20%75%72%6C%0A%20%20%20%20%20%20%20%20%64%6C%67%2E%67%72%2E%62%74%6E%73%47%72%2E%76%69%73%69%74%42%74%6E%2E%6F%6E%43%6C%69%63%6B%20%3D%20%66%75%6E%63%74%69%6F%6E%20%28%29%20%0A%20%20%20%20%20%20%20%20%7B%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%63%6D%64%20%3D%20%22%22%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%76%61%72%20%75%72%6C%20%3D%20%22%68%74%74%70%3A%2F%2F%77%77%77%2E%6E%61%62%73%63%72%69%70%74%73%2E%63%6F%6D%2F%22%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%69%66%20%28%24%2E%6F%73%2E%69%6E%64%65%78%4F%66%28%22%57%69%6E%22%29%20%21%3D%20%2D%31%29%0A%20%20%20%20%20%20%20%20%20%20%20%20%7B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%69%66%20%28%46%69%6C%65%28%22%43%3A%2F%50%72%6F%67%72%61%6D%20%46%69%6C%65%73%2F%4D%6F%7A%69%6C%6C%61%20%46%69%72%65%66%6F%78%2F%66%69%72%65%66%6F%78%2E%65%78%65%22%29%2E%65%78%69%73%74%73%29%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%7B%0A%09%09%09%09%09%63%6D%64%20%2B%3D%20%22%43%3A%2F%50%72%6F%67%72%61%6D%20%46%69%6C%65%73%2F%4D%6F%7A%69%6C%6C%61%20%46%69%72%65%66%6F%78%2F%66%69%72%65%66%6F%78%2E%65%78%65%20%22%20%2B%20%75%72%6C%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%7D%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%65%6C%73%65%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%7B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%09%63%6D%64%20%2B%3D%20%22%43%3A%2F%50%72%6F%67%72%61%6D%20%46%69%6C%65%73%2F%49%6E%74%65%72%6E%65%74%20%45%78%70%6C%6F%72%65%72%2F%69%65%78%70%6C%6F%72%65%2E%65%78%65%20%22%20%2B%20%75%72%6C%3B%0A%09%09%09%09%7D%0A%20%20%20%20%20%20%20%20%20%20%20%20%7D%0A%20%20%20%20%20%20%20%20%20%20%20%20%65%6C%73%65%0A%20%20%20%20%20%20%20%20%20%20%20%20%7B%0A%20%20%20%20%20%20%20%20%20%20%20%20%09%63%6D%64%20%2B%3D%20%22%6F%70%65%6E%20%5C%22%22%20%2B%20%75%72%6C%20%2B%20%22%5C%22%22%3B%0A%20%20%20%20%20%20%20%20%20%20%20%20%7D%0A%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%74%72%79%0A%20%20%20%20%20%20%20%20%20%20%20%20%7B%0A%09%09%09%09%73%79%73%74%65%6D%2E%63%61%6C%6C%53%79%73%74%65%6D%28%63%6D%64%29%3B%0A%09%09%09%7D%0A%09%09%09%63%61%74%63%68%28%65%29%0A%09%09%09%7B%0A%09%09%09%09%61%6C%65%72%74%28%65%29%3B%0A%09%09%09%7D%0A%20%20%20%20%20%20%20%20%7D%0A%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%64%6C%67%2E%63%65%6E%74%65%72%28%29%3B%0A%20%20%20%20%20%20%20%20%64%6C%67%2E%73%68%6F%77%28%29%3B'));
    };
}


/**
 Creates an instance of the main class and run it
*/
new Create3DBox().run(this);
